"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_extra_1 = __importStar(require("fs-extra"));
const path_1 = __importDefault(require("path"));
const execa_1 = __importDefault(require("execa"));
const build_utils_1 = require("@now/build-utils");
const dev_1 = require("./dev");
const util_1 = require("./util");
const DENO_LATEST = 'latest';
const DENO_VERSION = process.env.DENO_VERSION || DENO_LATEST;
const DOWNLOAD_URL = DENO_VERSION === DENO_LATEST ?
    `https://github.com/hayd/deno-lambda/releases/latest/download/deno-lambda-layer.zip` :
    `https://github.com/hayd/deno-lambda/releases/download/${DENO_VERSION}/deno-lambda-layer.zip`;
build_utils_1.debug('Deno Version:', DENO_VERSION);
exports.version = 3;
async function build(opts) {
    const { files, entrypoint, workPath: wp, config, meta = {} } = opts;
    const workPath = util_1.getWorkPath(wp, entrypoint);
    await fs_extra_1.default.mkdirp(workPath);
    if (meta.isDev) {
        build_utils_1.debug('checking that deno is available');
        dev_1.ensureDeno();
        build_utils_1.debug('checking that bash is available');
        dev_1.ensureBash();
    }
    const lambdaFiles = await getDenoLambdaLayer(workPath, meta.isDev || false);
    if (meta.isDev) {
        build_utils_1.debug('symlinking local deno to replace deno-lambda-layer bin/deno');
        await dev_1.replaceBinDeno(workPath);
    }
    build_utils_1.debug('downloading source files');
    const downloadedFiles = await build_utils_1.download(files, path_1.default.join(workPath, 'src'), meta);
    const entryPath = downloadedFiles[entrypoint].fsPath;
    await runUserScripts(entryPath);
    const extraFiles = await gatherExtraFiles(config.includeFiles, entryPath);
    return buildDenoLambda(opts, downloadedFiles, extraFiles, lambdaFiles, workPath);
}
exports.build = build;
async function buildDenoLambda({ entrypoint }, downloadedFiles, extraFiles, layerFiles, workPath) {
    const unstable = !!process.env.DENO_UNSTABLE;
    build_utils_1.debug('building single file');
    const entrypointPath = downloadedFiles[entrypoint].fsPath;
    const entrypointDirname = path_1.default.dirname(entrypointPath);
    const extname = path_1.default.extname(entrypointPath);
    const binName = path_1.default.basename(entrypointPath).replace(extname, '');
    const binPath = path_1.default.join(workPath, binName) + '.bundle.js';
    const denoDir = path_1.default.join(workPath, 'layer', '.deno_dir');
    build_utils_1.debug('running `deno bundle`...');
    try {
        await execa_1.default(path_1.default.join(workPath, 'layer', 'bin', 'deno'), ['bundle', entrypointPath, binPath, ...(unstable ? ['--unstable'] : [])], {
            env: {
                DENO_DIR: denoDir,
            },
            cwd: entrypointDirname,
            stdio: 'pipe',
        });
    }
    catch (err) {
        build_utils_1.debug('failed to `deno bundle`');
        throw new Error('Failed to run `deno bundle`: ' + err.stdout + ' ' + err.stderr);
    }
    const denoDirFiles = await getDenoDirFiles(denoDir);
    const lambda = await build_utils_1.createLambda({
        files: {
            ...extraFiles,
            ...layerFiles,
            ...denoDirFiles,
            [binName + '.bundle.js']: new build_utils_1.FileFsRef({
                mode: 0o755,
                fsPath: binPath,
            }),
        },
        handler: binName + '.handler',
        runtime: 'provided',
        environment: {
            HANDLER_EXT: 'bundle.js',
            PATH: process.env.PATH + ':./bin',
        },
    });
    if (exports.version === 3) {
        return { output: lambda };
    }
    return {
        [entrypoint]: lambda,
    };
}
async function walk(dir) {
    const f = await fs_extra_1.default.readdir(dir);
    const files = await Promise.all(f.map(async (file) => {
        const filePath = path_1.default.join(dir, file);
        const stats = await fs_extra_1.default.stat(filePath);
        if (stats.isDirectory())
            return walk(filePath);
        else if (stats.isFile())
            return filePath;
        throw 'File not dir or file: ' + filePath;
    }));
    return files.flat();
}
async function getDenoDirFiles(denoDirPath) {
    const files = {};
    const dir = await walk(denoDirPath);
    dir.forEach(file => {
        const f = path_1.default.join('.deno_dir', file.replace(denoDirPath + '/', ''));
        files[f] = new build_utils_1.FileFsRef({ fsPath: file, mode: 0o755 });
    });
    return files;
}
async function getDenoLambdaLayer(workPath, isDev) {
    const zipPath = path_1.default.join(workPath, 'deno-lambda-layer.zip');
    if (!(await fs_extra_1.pathExists(zipPath))) {
        build_utils_1.debug('downloading ', DOWNLOAD_URL);
        try {
            await execa_1.default('curl', [
                '-o',
                zipPath,
                '-L',
                DOWNLOAD_URL,
            ], {
                stdio: 'pipe',
            });
        }
        catch (err) {
            build_utils_1.debug('failed to download deno-lambda-layer');
            throw new Error('Failed to download deno-lambda-layer.zip: ' +
                err.stdout +
                ' ' +
                err.stderr);
        }
    }
    const layerDir = path_1.default.join(workPath, 'layer');
    const bootstrapPath = path_1.default.join(layerDir, 'bootstrap');
    const denoPath = path_1.default.join(layerDir, 'bin/deno');
    if (!(await fs_extra_1.pathExists(bootstrapPath)) || !(await fs_extra_1.pathExists(denoPath))) {
        build_utils_1.debug('unzipping `deno-lambda-layer.zip` into `layer`');
        try {
            await execa_1.default('unzip', [zipPath, '-d', layerDir], {
                stdio: 'pipe',
            });
        }
        catch (err) {
            build_utils_1.debug('failed to unzip `deno-lambda-layer.zip` into `layer`');
            throw new Error('Failed to unzip `deno-lambda-layer.zip` into `layer`: ' +
                err.stdout +
                ' ' +
                err.stderr);
        }
    }
    if (isDev) {
        build_utils_1.debug('using bash for bootstrap script');
        await dev_1.replaceBootstrapBash(bootstrapPath);
    }
    return {
        bootstrap: new build_utils_1.FileFsRef({
            mode: 0o755,
            fsPath: bootstrapPath,
        }),
        'bin/deno': new build_utils_1.FileFsRef({
            mode: 0o755,
            fsPath: denoPath,
        }),
    };
}
async function runUserScripts(entrypoint) {
    const entryDir = path_1.default.dirname(entrypoint);
    const buildScriptPath = path_1.default.join(entryDir, 'build.sh');
    const buildScriptExists = await fs_extra_1.default.pathExists(buildScriptPath);
    if (buildScriptExists) {
        console.log('running `build.sh`...');
        await build_utils_1.runShellScript(buildScriptPath);
    }
}
async function gatherExtraFiles(globMatcher, entrypoint) {
    if (!globMatcher)
        return {};
    console.log('gathering extra files for the fs...');
    const entryDir = path_1.default.dirname(entrypoint);
    if (Array.isArray(globMatcher)) {
        const allMatches = await Promise.all(globMatcher.map(pattern => build_utils_1.glob(pattern, entryDir)));
        return allMatches.reduce((acc, matches) => ({ ...acc, ...matches }), {});
    }
    return build_utils_1.glob(globMatcher, entryDir);
}
